package ui.controller;

import java.awt.Color;
import java.awt.Point;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.ListIterator;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;
import com.google.gson.stream.JsonWriter;

import TypeAdapter.AbstractCpsObjectAdapter;
import TypeAdapter.ColorAdapter;

import TypeAdapter.PositionAdapter;
import classes.Category;
import classes.CpsEdge;
import classes.CpsUpperNode;
import classes.AbstractCpsObject;
import classes.HolonElement;
import classes.HolonObject;
import classes.HolonSwitch;
import classes.HolonTransformer;
import classes.IdCounter;
import classes.Position;
import sun.misc.Queue;
import ui.model.Model;

/**
 * Controller for Storage.
 * 
 * @author Gruppe14
 */
public class StoreController {

	public enum MODE {
		COMPLETE, PARTIAL
	}

	public enum TYPE {
		CATEGORY, CANVAS
	}

	public enum EDGETYPE {
		CANVAS, CONNECTION, NODE, OLD
	}

	public enum NUMTYPE {
		CATEGORY, OBJECT, ELEMENT, EDGE, CONNECTION, NODEEDGE, OLDEDGE
	}

	private Model model;
	private int nCat, nObj, nEle, nEdge, nConn, nNodeEdge, nOldEdge;

	/**
	 * Constructor.
	 * 
	 * @param model
	 *            the Model
	 */
	public StoreController(Model model) {
		this.model = model;
		initNumeration();
	}

	/**
	 * Writes the current State of the Modelling into a JSON File which can be
	 * loaded.
	 * 
	 * @param path
	 *            the Path
	 * 
	 * @throws IOException
	 *             exception
	 */
	public void writeSaveFile(String path) throws IOException {

		// instance of the builder and give the wanted options
		GsonBuilder builder = new GsonBuilder();
		builder.excludeFieldsWithoutExposeAnnotation().serializeNulls();
		builder.registerTypeAdapter(AbstractCpsObject.class, new AbstractCpsObjectAdapter());
		builder.registerTypeAdapter(Position.class, new PositionAdapter());
		builder.registerTypeAdapter(Color.class, new ColorAdapter());
		// use the builder and make a instance of the Gson
		Gson gson = builder.create();

		JsonObject file = new JsonObject();

		initialize(MODE.COMPLETE, file);
		storeCategory(gson, file);
		storeCanvas(gson, file);
		System.out.println(file.toString());
		// writeCanvasObjects(json);
		// writeCategoryElements(json);
		// writeCanvasElements(json);
		// writeEdges(json);
		// writeElementGraph(json);
		int i = 1;
		// for (AbstractCpsObject cps : model.getObjectsOnCanvas()) {
		// String k = objectGson.toJson(cps, AbstractCpsObject.class);
		// System.out.println(k);
		// }

		// for (CpsEdge edge : model.getEdgesOnCanvas()) {
		//
		// JsonObject k = new JsonObject();
		// k.add("Edge", objectGson.toJsonTree(edge));
		// k.add("a", new JsonPrimitive(edge.getA().getID()));
		// k.add("b", new JsonPrimitive(edge.getB().getID()));
		// System.out.println(k);
		// }

		FileWriter writer = new FileWriter(path);
		// writer.write(json.toJSONString());
		getClass();

		writer.flush();
		writer.close();
	}

	/**
	 * Write needed default parameter into the JsonObject. Can be extended later
	 * on
	 * 
	 * @param mode
	 * @param file
	 */
	private void initialize(MODE mode, JsonObject file) {
		// TODO Auto-generated method stub
		switch (mode) {
		case COMPLETE:
			file.add("MODE", new JsonPrimitive(mode.name()));
			file.add("IDCounter", new JsonPrimitive(IdCounter.getCounter()));
			file.add("CANVAS_SIZE_X", new JsonPrimitive(model.getCanvasX()));
			file.add("CANVAS_SIZE_Y", new JsonPrimitive(model.getCanvasY()));
			break;
		case PARTIAL:
			file.add("MODE", new JsonPrimitive(mode.name()));
			file.add("IDCounter", new JsonPrimitive(IdCounter.getCounter()));
			break;

		default:
			break;
		}
	}

	/**
	 * Store all Categories and Object into a Json File via Serialization
	 * 
	 * @param gson
	 * @param file
	 */
	private void storeCategory(Gson gson, JsonObject file) {
		// TODO Auto-generated method stub
		for (Category cat : model.getCategories()) {
			String key = "CATEGORY" + getNumerator(NUMTYPE.CATEGORY);
			file.add(key, new JsonPrimitive(cat.getName()));
			for (AbstractCpsObject obj : cat.getObjects()) {
				file.add("CATEGORY.OBJECT" + getNumerator(NUMTYPE.OBJECT),
						new JsonPrimitive(gson.toJson(obj, AbstractCpsObject.class)));
				if (obj instanceof HolonObject)
					elementsToJson(TYPE.CATEGORY, gson, file, obj);
			}
		}
	}

	/**
	 * 
	 * @param gson
	 * @param file
	 */
	private void storeCanvas(Gson gson, JsonObject file) {
		// TODO Auto-generated method stub
		Queue<AbstractCpsObject> queue = new Queue<>();
		AbstractCpsObject u = null;

		edgeToJson(EDGETYPE.CANVAS, gson, file, 0, model.getEdgesOnCanvas());

		for (AbstractCpsObject cps : model.getObjectsOnCanvas()) {
			queue.enqueue(cps);
		}

		while (!queue.isEmpty()) {
			try {
				u = queue.dequeue();
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			String key = "CANVAS.OBJECT" + getNumerator(NUMTYPE.OBJECT);
			file.add(key, new JsonPrimitive(gson.toJson(u, AbstractCpsObject.class)));
			edgeToJson(EDGETYPE.CONNECTION, gson, file, u.getID(), u.getConnections());

			if (u instanceof HolonObject)
				elementsToJson(TYPE.CANVAS, gson, file, u);
			if (u instanceof CpsUpperNode) {
				for (AbstractCpsObject adjacent : ((CpsUpperNode) u).getNodes()) {
					queue.enqueue(adjacent);

				}
				edgeToJson(EDGETYPE.NODE, gson, file, u.getID(), ((CpsUpperNode) u).getNodeEdges());
				edgeToJson(EDGETYPE.OLD, gson, file, u.getID(), ((CpsUpperNode) u).getOldEdges());
			}
		}

	}

	/**
	 * Write the Canvas File.
	 * 
	 * @param path
	 *            the Path
	 * @throws IOException
	 *             Exception
	 */
	public void writeCanvasFile(String path) throws IOException {

		JSONObject json = new JSONObject();

		json.put("MODE", "CANVAS");
		json.put("ID", IdCounter.getCounter());
		writeCanvasObjects(json);
		writeCanvasElements(json);
		writeEdges(json);
		writeElementGraph(json);

		FileWriter writer = new FileWriter(path);
		writer.write(json.toJSONString());
		getClass();

		writer.flush();
		writer.close();
	}

	/**
	 * writes all Categories into a JSONObject.
	 * 
	 * @param gson
	 * 
	 * @param json
	 *            JSON Object
	 * @throws IOException
	 *             exception
	 */
	public void writeCategory(Gson gson, JSONObject json) {
		int i = 1;

		for (Category cat : model.getCategories())
			json.put("CG" + i++, gson.toJson(cat));
	}

	/**
	 * writes all Objects in Category into a JSONObject.
	 * 
	 * @param gson
	 * 
	 * @param json
	 *            JSON Object
	 */
	public void writeCategoryObjects(Gson gson, JSONObject json) {

		int i = 1;

		for (Category cats : model.getCategories())
			for (AbstractCpsObject cps : cats.getObjects()) {
				// arr.add(cps.getSav());
				// arr.add(cps.getObjName());
				// arr.add(cps.getImage());
				json.put("CGO" + i++, gson.toJson(cps));

			}
	}

	/**
	 * Wrte Canvas Objects.
	 * 
	 * @param json
	 *            JSON Object
	 */
	public void writeCanvasObjects(JSONObject json) {

		JSONArray arr = new JSONArray();
		int i = 1;
		for (AbstractCpsObject cps : model.getObjectsOnCanvas()) {
			// arr.add(getObjectType(cps));
			arr.add(cps.getObjName());
			arr.add(cps.getName());
			arr.add(cps.getID());
			arr.add(cps.getImage());
			arr.add(cps.getPosition().x);
			arr.add(cps.getPosition().y);
			json.put("CVSO" + i++, arr);
			arr = new JSONArray();
		}
	}

	/**
	 * writes all Elements in Objects in Category into a JSONObject.
	 * 
	 * @param json
	 *            JSON Object
	 */
	public void writeCategoryElements(JSONObject json) {

		JSONArray arr = new JSONArray();
		int i = 1;

		for (Category cats : model.getCategories())
			for (AbstractCpsObject cps : cats.getObjects())
				if (cps instanceof HolonObject)
					for (HolonElement ele : ((HolonObject) cps).getElements()) {
						arr.add(ele.getSav());
						arr.add(ele.getObj());
						arr.add(ele.getEleName());
						arr.add(ele.getAmount());
						arr.add(ele.getEnergy());
						json.put("CGE" + i++, arr);
						arr = new JSONArray();
					}

	}

	/**
	 * Write Canvas Elements.
	 * 
	 * @param json
	 *            JSON Objects
	 */
	public void writeCanvasElements(JSONObject json) {

		JSONArray arr = new JSONArray();
		int i = 1;
		for (AbstractCpsObject cps : model.getObjectsOnCanvas()) {
			if (cps instanceof HolonObject)
				for (HolonElement ele : ((HolonObject) cps).getElements()) {
					arr.add(ele.getSav());
					arr.add(cps.getID());
					arr.add(ele.getEleName());
					arr.add(ele.getAmount());
					arr.add(ele.getEnergy());
					arr.add(Boolean.compare(ele.getActive(), true) + 1);
					json.put("CVSE" + i++, arr);
					arr = new JSONArray();
				}
		}
	}

	/**
	 * write all Edges into a JSONObject.
	 * 
	 * @param json
	 *            JSON Object
	 */
	public void writeEdges(JSONObject json) {

		JSONArray arr = new JSONArray();
		int i = 1;

		for (CpsEdge edge : model.getEdgesOnCanvas()) {
			arr.add(edge.getA().getID());
			arr.add(edge.getB().getID());
			arr.add(edge.getCapacity());
			arr.add(edge.getFlow());
			json.put("EDGE" + i++, arr);
			arr = new JSONArray();
		}
	}

	/**
	 * writes the Graph from all Elements into a JSONObject.
	 * 
	 * @param json
	 *            JSON Object
	 */
	public void writeElementGraph(JSONObject json) {

		ListIterator<Point> iterator;
		JSONArray arr = new JSONArray();
		int i = 1;

		for (Category cats : model.getCategories())
			for (AbstractCpsObject cps : cats.getObjects())
				if (cps instanceof HolonObject)
					for (HolonElement ele : ((HolonObject) cps).getElements())
						if (!ele.getGraphPoints().isEmpty()) {
							iterator = ele.getGraphPoints().listIterator();

							arr.add(ele.getSav());
							arr.add(ele.getObj());
							arr.add(ele.getEleName());

							while (iterator.hasNext()) {
								Point p = iterator.next();
								arr.add((int) p.getX());
								arr.add((int) p.getY());
							}
							json.put("CGGP" + i++, arr);
							arr = new JSONArray();
						}
		i = 1;
		for (AbstractCpsObject cps : model.getObjectsOnCanvas()) {
			if (cps instanceof HolonObject)
				for (HolonElement ele : ((HolonObject) cps).getElements())
					if (!ele.getGraphPoints().isEmpty()) {
						iterator = ele.getGraphPoints().listIterator();

						arr.add(ele.getSav());
						arr.add(cps.getID());
						arr.add(ele.getEleName());

						while (iterator.hasNext()) {
							Point p = iterator.next();
							arr.add((int) p.getX());
							arr.add((int) p.getY());
						}
						json.put("CVSGP" + i++, arr);
						arr = new JSONArray();
					}
		}
	}

	private void elementsToJson(TYPE type, Gson gson, JsonObject file, AbstractCpsObject obj) {
		// TODO Auto-generated method stub
		JsonObject temp = new JsonObject();
		
		for (HolonElement ele : ((HolonObject) obj).getElements()) {
			file.add(type.name() + ".OBJECT.ELEMENT" + +getNumerator(NUMTYPE.ELEMENT),
					new JsonPrimitive(gson.toJson(ele)));
		}

	}

	private void edgeToJson(EDGETYPE type, Gson gson, JsonObject file, int id, ArrayList<CpsEdge> arr) {
		// TODO Auto-generated method stub
		String k = null;
		int i = 0;
		JsonObject temp = new JsonObject();

		for (CpsEdge edge : arr) {

			temp.add("properties", new JsonPrimitive(gson.toJson(edge)));
			temp.add("A", new JsonPrimitive(edge.getA().getID()));
			temp.add("B", new JsonPrimitive(edge.getB().getID()));

			switch (type) {
			case CANVAS:
				k = "CANVAS.EDGE" + getNumerator(NUMTYPE.EDGE);
				break;
			case CONNECTION:
				k = "CANVAS.OBJECT.CONNECTION" + getNumerator(NUMTYPE.CONNECTION);
				break;
			case NODE:
				k = "CANVAS.UPPERNODE.EDGE" + getNumerator(NUMTYPE.NODEEDGE);
				temp.add("ID", new JsonPrimitive(id));
				break;
			case OLD:
				k = "CANVAS.UPPERNODE.OLD" + getNumerator(NUMTYPE.OLDEDGE);
				temp.add("ID", new JsonPrimitive(id));
				break;
			default:
				break;
			}

			file.add(k, gson.toJsonTree(temp));
			temp = new JsonObject();
		}

	}

	/**
	 * 
	 * @param type
	 * @return
	 */
	private int getNumerator(NUMTYPE type) {

		switch (type) {
		case CATEGORY:
			return nCat++;
		case OBJECT:
			return nObj++;
		case ELEMENT:
			return nEle++;
		case EDGE:
			return nEdge++;
		case CONNECTION:
			return nConn++;
		case NODEEDGE:
			return nNodeEdge++;
		case OLDEDGE:
			return nOldEdge++;
		default:
			break;
		}
		return -1;
	}

	/**
	 * Just initialize the Numerators for the Json Keys. Maybe bad Style..
	 */
	private void initNumeration() {
		this.nCat = this.nObj = this.nEle = this.nEdge = this.nConn = this.nNodeEdge = this.nOldEdge = 0;
	}

}